import json
import logging
import os

import tensorflow as tf

try:
    AUTOTUNE = tf.data.AUTOTUNE
except:
    AUTOTUNE = tf.data.experimental.AUTOTUNE


def read_tfrecord_files(input_files, num_parallel_calls=None, **kwargs):
    num_parallel_calls = num_parallel_calls or AUTOTUNE
    if isinstance(input_files, str):
        input_files = [input_files]
    if len(input_files) == 1:
        dataset = tf.data.TFRecordDataset(input_files)
    else:
        dataset = tf.data.Dataset.from_tensor_slices(input_files)
        dataset = dataset.interleave(
            lambda x: tf.data.TFRecordDataset(x),
            cycle_length=len(input_files),
            num_parallel_calls=num_parallel_calls,
        )
    return dataset


def batching_and_padding(dataset: tf.data.Dataset, padded_shapes, padding_values, **kwargs) -> tf.data.Dataset:
    dataset = dataset.padded_batch(
        batch_size=kwargs.get("batch_size", 32),
        padded_shapes=padded_shapes,
        padding_values=padding_values,
        drop_remainder=kwargs.get("drop_remainder", False),
    )
    return dataset


def bucketing_and_padding(dataset: tf.data.Dataset, bucket_fn, padded_shapes, padding_values, **kwargs) -> tf.data.Dataset:
    batch_size = kwargs.get("batch_size", 32)
    bucket_boundaries = kwargs.get("bucket_boundaries", None)
    if not bucket_boundaries:
        bucket_boundaries = []
        maxlen = kwargs.get("max_sequence_length", 512)
        numbuk = kwargs.get("num_buckets", 8)
        step = maxlen // numbuk
        for i in range(1, numbuk + 1):
            v = i * step
            if v >= maxlen:
                break
            bucket_boundaries.append(v)

    bucket_batch_sizes = kwargs.get("bucket_batch_sizes", None)
    if not bucket_batch_sizes:
        bucket_batch_sizes = [batch_size] * (1 + len(bucket_boundaries))

    assert len(bucket_batch_sizes) == len(bucket_boundaries) + 1, "len(bucket_batch_sizes) != len(bucket_boundaries) + 1"

    try:
        fn = tf.data.bucket_by_sequence_length
    except:
        fn = tf.data.experimental.bucket_by_sequence_length

    dataset = dataset.apply(
        fn(
            element_length_func=bucket_fn,
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=kwargs.get("drop_remainder", False),
            pad_to_bucket_boundary=kwargs.get("pad_to_bucket_boundary", False),
            no_padding=kwargs.get("no_padding", False),
        )
    )
    return dataset


def read_jsonl_files(input_files, select_keys=None, ignore_keys=None, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                d = json.loads(line)
                if select_keys:
                    for k in list(d.keys()):
                        if k not in select_keys:
                            d.pop(k, None)
                if ignore_keys:
                    for k in list(d.keys()):
                        if k in ignore_keys:
                            d.pop(k, None)
                yield d


def int64_feature(values):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


def save_tfrecord(examples, fn, output_files, **kwargs):
    if not examples:
        logging.warning("Examples is empty or None, skipped.")
        return
    if isinstance(output_files, str):
        output_files = [output_files]
    writers = [tf.io.TFRecordWriter(f) for f in output_files]
    idx, c = 0, 0
    for example in examples:
        if not example:
            continue
        try:
            feature = fn(example, **kwargs)
            record = tf.train.Example(features=tf.train.Features(feature=feature))
            writers[idx].write(record.SerializeToString())
        except Exception as e:
            logging.warning("Encode feature exception: ", e)
            continue
        c += 1
        idx += 1
        idx = idx % len(writers)
    for w in writers:
        w.close()
    logging.info("Finished to write %d examples to tfrecords.", c)



import unittest

from smile_datasets.mlm import readers
from smile_datasets.mlm.dataset import DatapipeForMaksedLanguageModel, DatasetForMaskedLanguageModel
from smile_datasets.mlm.example import ExampleForMaskedLanguageModel
from smile_datasets.mlm.parsers import ParserForMaskedLanguageModel


class MyDatasetForMaskedLanguageModel(DatasetForMaskedLanguageModel):
    """ """

    def __init__(self, input_files, vocab_file, **kwargs) -> None:
        super().__init__()
        examples = []
        parser = ParserForMaskedLanguageModel.from_vocab_file(vocab_file, **kwargs)
        for instance in readers.read_jsonl_files(input_files, **kwargs):
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        return self.examples[index]


class DatasetTest(unittest.TestCase):
    """ """

    def test_dataset_from_jsonl_files(self):
        d = MyDatasetForMaskedLanguageModel("testdata/mlm.jsonl", vocab_file="testdata/vocab.txt")
        print("Showing examples:")
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForMaksedLanguageModel.from_dataset(d)
        print(next(iter(dataset)))

        print("Save to tfrecord:")
        d.save_tfrecord("testdata/mlm.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForMaksedLanguageModel.from_tfrecord_files("testdata/mlm.tfrecord")
        print(next(iter(dataset)))

    def test_datapipe_from_jsonl_files(self):
        print()
        dataset = DatapipeForMaksedLanguageModel.from_jsonl_files("testdata/mlm.jsonl", vocab_file="testdata/vocab.txt")
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from smile_datasets.question_answering import readers
from smile_datasets.question_answering.dataset import DatapipeForQuestionAnswering, DatasetForQuestionAnswering
from smile_datasets.question_answering.example import ExampleForQuestionAnswering
from smile_datasets.question_answering.parsers import ParserForQuestionAnswering

DUREADER_RUBOST_INPUT_FILE = os.path.join(os.environ["DUREADER_ROBUST_PATH"], "dev.json")
DUREADER_CHECKLIST_INPUT_FILE = os.path.join(os.environ["DUREADER_CHECKLIST_PATH"], "dev.json")
VOCAB_FILE = os.environ["BERT_VOCAB_PATH"]


class DuReaderDatasetForQuestionAnswering(DatasetForQuestionAnswering):
    """ """

    def __init__(self, input_files, vocab_file, subset="rubost", **kwargs) -> None:
        super().__init__()
        self.parser = ParserForQuestionAnswering(tokenizer=None, vocab_file=vocab_file, **kwargs)
        if subset == "rubost":
            self.instances = list(readers.read_dureader_rubost(input_files, **kwargs))
        else:
            self.instances = list(readers.read_dureader_checklist(input_files, **kwargs))
        self.examples = []
        for instance in self.instances:
            e = self.parser.parse(instance)
            if not e:
                continue
            self.examples.append(e)

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index) -> ExampleForQuestionAnswering:
        return self.examples[index]


class DatasetTest(unittest.TestCase):
    """ """

    def test_dataset_save_dureader_rubost_tfrecord(self):
        d = DuReaderDatasetForQuestionAnswering(DUREADER_RUBOST_INPUT_FILE, VOCAB_FILE)
        print("\nShowing examples: ")
        for idx, e in enumerate(d):
            print(e)
            if idx == 5:
                break

        print("Load datapipe from dataset:")
        dataset = DatapipeForQuestionAnswering.from_dataset(d)
        print()
        print(next(iter(dataset)))

        print("Save examples to tfrecord:")
        d.save_tfrecord("testdata/dureader_rubost_dev.tfrecord")

        print("Load datapipe from tfrecord files:")
        dataset = DatapipeForQuestionAnswering.from_tfrecord_files("testdata/dureader_rubost_dev.tfrecord")
        print()
        print(next(iter(dataset)))

    def test_dataset_save_dureader_checklist_tfrecord(self):
        d = DuReaderDatasetForQuestionAnswering(DUREADER_CHECKLIST_INPUT_FILE, VOCAB_FILE, subset="checklist")
        print("\nShowing examples: ")
        for idx, e in enumerate(d):
            print(e)
            if idx == 5:
                break

        print("Load datapipe from dataset:")
        dataset = DatapipeForQuestionAnswering.from_dataset(d)
        print()
        print(next(iter(dataset)))

        print("Save examples to tfrecord:")
        d.save_tfrecord("testdata/dureader_checklist_dev.tfrecord")

        print("Load datapipe from tfrecord files:")
        dataset = DatapipeForQuestionAnswering.from_tfrecord_files("testdata/dureader_checklist_dev.tfrecord")
        print()
        print(next(iter(dataset)))

    def test_datapipe_from_dureader_rubost(self):
        dataset = DatapipeForQuestionAnswering.from_dureader_robust(DUREADER_RUBOST_INPUT_FILE, vocab_file=VOCAB_FILE)
        print()
        print(next(iter(dataset)))

    def test_datapipe_from_dureader_checklist(self):
        dataset = DatapipeForQuestionAnswering.from_dureader_checklist(DUREADER_CHECKLIST_INPUT_FILE, vocab_file=VOCAB_FILE)
        print()
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import unittest

from smile_datasets.simcse import readers
from smile_datasets.simcse.example import ExampleForHardNegativeSimCSE, ExampleForSupervisedSimCSE, ExampleForUnsupervisedSimCSE
from smile_datasets.simcse.hardneg_dataset import DatapipeForHardNegativeSimCSE, DatasetForHardNegativeSimCSE
from smile_datasets.simcse.parsers import ParserForHardNegSimCSE, ParserForSupervisedSimCSE, ParserForUnsupervisedSimCSE
from smile_datasets.simcse.sup_dataset import DatapipeForSupervisedSimCSE, DatasetForSupervisedSimCSE
from smile_datasets.simcse.unsup_dataset import DatapipeForUnsupervisedSimCSE, DatasetForUnsupervisedSimCSE


class MyDatasetForUnsupSimCSE(DatasetForUnsupervisedSimCSE):
    """ """

    def __init__(self, input_files, vocab_file, **kwargs) -> None:
        super().__init__()
        examples = []
        parser = ParserForUnsupervisedSimCSE.from_vocab_file(vocab_file, **kwargs)
        for instance in readers.read_jsonl_files_for_unsup_simcse(input_files, **kwargs):
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index) -> ExampleForUnsupervisedSimCSE:
        return self.examples[index]


class MyDatasetForSupervisedSimCSE(DatasetForSupervisedSimCSE):
    """ """

    def __init__(self, input_files, vocab_file, **kwargs) -> None:
        super().__init__()
        examples = []
        parser = ParserForSupervisedSimCSE.from_vocab_file(vocab_file, **kwargs)
        for instance in readers.read_jsonl_files_for_supervised_simcse(input_files, **kwargs):
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index) -> ExampleForSupervisedSimCSE:
        return self.examples[index]


class MyDatasetForHardNegativeSimCSE(DatasetForHardNegativeSimCSE):
    """ """

    def __init__(self, input_files, vocab_file, **kwargs) -> None:
        super().__init__()
        examples = []
        parser = ParserForHardNegSimCSE.from_vocab_file(vocab_file, **kwargs)
        for instance in readers.read_jsonl_files_for_hardneg_simcse(input_files, **kwargs):
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index) -> ExampleForHardNegativeSimCSE:
        return self.examples[index]


class DatasetTest(unittest.TestCase):
    """ """

    def test_unsup_dataset_from_jsonl(self):
        d = MyDatasetForUnsupSimCSE("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print()
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForUnsupervisedSimCSE.from_dataset(d)
        print(next(iter(dataset)))

        print("Save tfrecord:")
        d.save_tfrecord("testdata/simcse.unsup.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForUnsupervisedSimCSE.from_tfrecord_files("testdata/simcse.unsup.tfrecord")
        print(next(iter(dataset)))

    def test_sup_datapipe_from_jsonl(self):
        d = MyDatasetForSupervisedSimCSE("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print()
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForSupervisedSimCSE.from_dataset(d)
        print(next(iter(dataset)))

        print("Save tfrecord:")
        d.save_tfrecord("testdata/simcse.sup.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForSupervisedSimCSE.from_tfrecord_files("testdata/simcse.sup.tfrecord")
        print(next(iter(dataset)))

    def test_hardnef_dataset_from_jsonl(self):
        d = MyDatasetForHardNegativeSimCSE("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print()
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForHardNegativeSimCSE.from_dataset(d)
        print(next(iter(dataset)))

        print("Save tfrecord:")
        d.save_tfrecord("testdata/simcse.hardneg.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForHardNegativeSimCSE.from_tfrecord_files("testdata/simcse.hardneg.tfrecord")
        print(next(iter(dataset)))

    def test_unsup_datapipe_from_jsonl(self):
        print()
        dataset = DatapipeForUnsupervisedSimCSE.from_jsonl_files("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print(next(iter(dataset)))

    def test_sup_dataset_from_jsonl(self):
        print()
        dataset = DatapipeForSupervisedSimCSE.from_jsonl_files("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print(next(iter(dataset)))

    def test_hardneg_datapipe_from_jsonl(self):
        print()
        dataset = DatapipeForHardNegativeSimCSE.from_jsonl_files("testdata/simcse.jsonl", vocab_file="testdata/vocab.txt")
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import logging

from smile_datasets.dataset import Datapipe, Dataset
from smile_datasets.mlm.dataset import DatapipeForMaksedLanguageModel, DatasetForMaskedLanguageModel
from smile_datasets.mlm.example import ExampleForMaskedLanguageModel
from smile_datasets.mlm.masking_strategy import WholeWordMask
from smile_datasets.mlm.parsers import ParserForMaskedLanguageModel
from smile_datasets.question_answering.dataset import DatapipeForQuestionAnswering, DatasetForQuestionAnswering
from smile_datasets.question_answering.example import ExampleForQuestionAnswering
from smile_datasets.question_answering.parsers import ParserForQuestionAnswering
from smile_datasets.sequence_classification.dataset import DatapipeForSequenceClassifiation, DatasetForSequenceClassification
from smile_datasets.sequence_classification.example import ExampleForSequenceClassification
from smile_datasets.sequence_classification.parsers import ParserForSequenceClassification
from smile_datasets.simcse.example import ExampleForHardNegativeSimCSE, ExampleForSupervisedSimCSE, ExampleForUnsupervisedSimCSE
from smile_datasets.simcse.hardneg_dataset import DatapipeForHardNegativeSimCSE, DatasetForHardNegativeSimCSE
from smile_datasets.simcse.parsers import ParserForHardNegSimCSE, ParserForSupervisedSimCSE, ParserForUnsupervisedSimCSE
from smile_datasets.simcse.sup_dataset import DatapipeForSupervisedSimCSE, DatasetForSupervisedSimCSE
from smile_datasets.simcse.unsup_dataset import DatapipeForUnsupervisedSimCSE, DatasetForUnsupervisedSimCSE
from smile_datasets.token_classification.dataset import DatapipeForTokenClassification, DatasetForTokenClassification
from smile_datasets.token_classification.example import ExampleForTokenClassification
from smile_datasets.token_classification.parsers import ParserForTokenClassification
from smile_datasets.token_classification.tokenizers import BertCharLevelTokenizer, LabelTokenizerForTokenClassification

__name__ = "smile_datasets"
__version__ = "0.0.6"

logging.basicConfig(format="%(asctime)s %(levelname)7s %(filename)20s %(lineno)4d] %(message)s", level=logging.INFO)



import json
import logging
import os


def read_jsonl_files_for_unsup_simcse(input_files, sequence_key="sequence", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning("File %d does not exist, skipped.")
            continue
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {"sequence": data[sequence_key]}
                yield instance


def read_jsonl_files_for_supervised_simcse(input_files, sequence_key="sequence", pos_sequence_key="pos_sequence", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning("File %d does not exist, skipped.")
            continue
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {"sequence": data[sequence_key], "pos_sequence": data[pos_sequence_key]}
                yield instance


def read_jsonl_files_for_hardneg_simcse(
    input_files, sequence_key="sequence", pos_sequence_key="pos_sequence", neg_sequence_key="neg_sequence", **kwargs
):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning("File %d does not exist, skipped.")
            continue
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {
                    "sequence": data[sequence_key],
                    "pos_sequence": data[pos_sequence_key],
                    "neg_sequence": data[neg_sequence_key],
                }
                yield instance



from collections import namedtuple

ExampleForMaskedLanguageModel = namedtuple(
    "ExampleForMaskedLanguageModel",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "masked_ids", "masked_pos"],
)



from tokenizers import BertWordPieceTokenizer

from .example import ExampleForHardNegativeSimCSE, ExampleForSupervisedSimCSE, ExampleForUnsupervisedSimCSE


class ParserForUnsupervisedSimCSE:
    """Parser for simcse"""

    @classmethod
    def from_tokenizer(cls, tokenizer: BertWordPieceTokenizer, **kwargs):
        return cls(tokenizer=tokenizer, vocab_file=None, **kwargs)

    @classmethod
    def from_vocab_file(cls, vocab_file, **kwargs):
        return cls(tokenizer=None, vocab_file=vocab_file, **kwargs)

    def __init__(
        self, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, do_lower_case=True, add_special_tokens=True, **kwargs
    ) -> None:
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided!"
        self.tokenizer = tokenizer or BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        self.add_special_tokens = add_special_tokens

    def parse(self, instance, **kwargs) -> ExampleForUnsupervisedSimCSE:
        sequence = instance["sequence"]
        encoding = self.tokenizer.encode(sequence, add_special_tokens=self.add_special_tokens)
        example = ExampleForUnsupervisedSimCSE(
            tokens=encoding.tokens,
            input_ids=encoding.ids,
            segment_ids=encoding.type_ids,
            attention_mask=encoding.attention_mask,
        )
        return example


class ParserForSupervisedSimCSE(ParserForUnsupervisedSimCSE):
    """Parser for supervised simcse."""

    def parse(self, instance, **kwargs) -> ExampleForSupervisedSimCSE:
        sequence, pos_sequence = instance["sequence"], instance["pos_sequence"]
        seq_encoding = self.tokenizer.encode(sequence, add_special_tokens=self.add_special_tokens)
        pos_encoding = self.tokenizer.encode(pos_sequence, add_special_tokens=self.add_special_tokens)
        example = ExampleForSupervisedSimCSE(
            tokens=seq_encoding.tokens,
            input_ids=seq_encoding.ids,
            segment_ids=seq_encoding.type_ids,
            attention_mask=seq_encoding.attention_mask,
            pos_tokens=pos_encoding.tokens,
            pos_input_ids=pos_encoding.ids,
            pos_segment_ids=pos_encoding.type_ids,
            pos_attention_mask=pos_encoding.attention_mask,
        )
        return example


class ParserForHardNegSimCSE(ParserForUnsupervisedSimCSE):
    """Parser for hard negative simcse."""

    def parse(self, instance, **kwargs) -> ExampleForHardNegativeSimCSE:
        sequence, pos_sequence, neg_sequence = instance["sequence"], instance["pos_sequence"], instance["neg_sequence"]
        seq_encoding = self.tokenizer.encode(sequence, add_special_tokens=self.add_special_tokens)
        pos_encoding = self.tokenizer.encode(pos_sequence, add_special_tokens=self.add_special_tokens)
        neg_encoding = self.tokenizer.encode(neg_sequence, add_special_tokens=self.add_special_tokens)
        example = ExampleForHardNegativeSimCSE(
            tokens=seq_encoding.tokens,
            input_ids=seq_encoding.ids,
            segment_ids=seq_encoding.type_ids,
            attention_mask=seq_encoding.attention_mask,
            pos_tokens=pos_encoding.tokens,
            pos_input_ids=pos_encoding.ids,
            pos_segment_ids=pos_encoding.type_ids,
            pos_attention_mask=pos_encoding.attention_mask,
            neg_tokens=neg_encoding.tokens,
            neg_input_ids=neg_encoding.ids,
            neg_segment_ids=neg_encoding.type_ids,
            neg_attention_mask=neg_encoding.attention_mask,
        )
        return example



import os
import unittest

from smile_datasets.sequence_classification import readers
from smile_datasets.sequence_classification.dataset import DatapipeForSequenceClassifiation, DatasetForSequenceClassification
from smile_datasets.sequence_classification.example import ExampleForSequenceClassification
from smile_datasets.sequence_classification.parsers import ParserForSequenceClassification

VOCAB_FILE = os.environ["BERT_VOCAB_PATH"]


class MyDatasetForSequenceClassification(DatasetForSequenceClassification):
    """ """

    def __init__(self, input_files, vocab_file, **kwargs) -> None:
        super().__init__()
        self.parser = ParserForSequenceClassification.from_vocab_file(vocab_file=vocab_file, **kwargs)
        self.instances = readers.read_jsonl_files(input_files, **kwargs)
        self.examples = []
        for instance in self.instances:
            e = self.parser.parse(instance, **kwargs)
            if not e:
                continue
            self.examples.append(e)

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        return self.examples[index]


class DatasetTest(unittest.TestCase):
    """Tests for sequence classification datasets"""

    def test_dataset_for_sequence_classification(self):
        d = MyDatasetForSequenceClassification("testdata/sequence_classification.jsonl", vocab_file=VOCAB_FILE)
        print("\nShowing examples:")
        for idx, e in enumerate(d):
            print(e)
        print()

        print("Save examples to tfrecord:")
        d.save_tfrecord("testdata/sequence_classification.tfrecord")

        print("Load datapipe from dataset:")
        dataset = DatapipeForSequenceClassifiation.from_dataset(d)
        print(next(iter(dataset)))

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForSequenceClassifiation.from_tfrecord_files("testdata/sequence_classification.tfrecord")
        print(next(iter(dataset)))

    def test_dataset_for_seqpair_classification(self):
        d = MyDatasetForSequenceClassification("testdata/seqpair_classification.jsonl", vocab_file=VOCAB_FILE)
        print("\nShowing examples:")
        for idx, e in enumerate(d):
            print(e)
        print()

        print("Save examples to tfrecord:")
        d.save_tfrecord("testdata/seqpair_classification.tfrecord")

        print("Load datapipe from dataset:")
        dataset = DatapipeForSequenceClassifiation.from_dataset(d)
        print(next(iter(dataset)))

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForSequenceClassifiation.from_tfrecord_files("testdata/seqpair_classification.tfrecord")
        print(next(iter(dataset)))

    def test_datapipe_from_jsonl_files(self):
        print()
        print("Load datapipe from sequence classification:")
        dataset = DatapipeForSequenceClassifiation.from_jsonl_files(
            "testdata/sequence_classification.jsonl", vocab_file=VOCAB_FILE
        )
        print(next(iter(dataset)))

        print("Load datapipe from seqpair classification:")
        dataset = DatapipeForSequenceClassifiation.from_jsonl_files(
            "testdata/seqpair_classification.jsonl", vocab_file=VOCAB_FILE
        )
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



import json
import logging
import os


def read_jsonl_files(input_files, sequence_key="sequence", label_key="label", pair_key="pair", id_key="id", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {
                    "sequence": data[sequence_key],
                    "label": data[label_key],
                }
                if pair_key and pair_key in data:
                    instance.update({"pair": data[pair_key]})
                if id_key in data:
                    instance.update({"id": data[id_key]})
                yield instance



import unittest

from smile_datasets import dataset
from smile_datasets.token_classification import readers
from smile_datasets.token_classification.dataset import DatapipeForTokenClassification, DatasetForTokenClassification
from smile_datasets.token_classification.example import ExampleForTokenClassification
from smile_datasets.token_classification.parsers import ParserForTokenClassification
from tensorflow._api.v2 import data


class MyDatasetForTokenClassification(DatasetForTokenClassification):
    """ """

    def __init__(self, input_files, feature_vocab_file, label_vocab_file, file_format="jsonl", **kwargs) -> None:
        super().__init__()
        if file_format == "jsonl":
            self.instances = readers.read_jsonl_files(input_files, **kwargs)
        elif file_format == "conll":
            self.instances = readers.read_conll_files(input_files, **kwargs)

        self.parser = ParserForTokenClassification.from_vocab_file(feature_vocab_file, label_vocab_file, **kwargs)

        examples = []
        for instance in self.instances:
            e = self.parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        self.examples = examples

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index) -> ExampleForTokenClassification:
        return self.examples[index]


class DatasetTest(unittest.TestCase):
    """Dataset tests for token classification"""

    def test_dataset_from_jsonl(self):
        d = MyDatasetForTokenClassification(
            "testdata/token_classification.jsonl",
            feature_vocab_file="testdata/vocab.txt",
            label_vocab_file="testdata/token_classification_vocab.txt",
            file_format="jsonl",
        )
        print("\nShowing examples:")
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForTokenClassification.from_dataset(d)
        print(next(iter(dataset)))

        print("Examples to tfrecord:")
        d.save_tfrecord("testdata/token_classification.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForTokenClassification.from_tfrecord_files("testdata/token_classification.tfrecord")
        print(next(iter(dataset)))

    def test_dataset_from_conll(self):
        d = MyDatasetForTokenClassification(
            "testdata/token_classification.conll.txt",
            feature_vocab_file="testdata/vocab.txt",
            label_vocab_file="testdata/token_classification_vocab.txt",
            file_format="conll",
        )
        print("\nShowing examples:")
        for _, e in enumerate(d):
            print(e)

        print("Load datapipe from dataset:")
        dataset = DatapipeForTokenClassification.from_dataset(d)
        print(next(iter(dataset)))

        print("Examples to tfrecord:")
        d.save_tfrecord("testdata/token_classification.conll.tfrecord")

        print("Load datapipe from tfrecord:")
        dataset = DatapipeForTokenClassification.from_tfrecord_files("testdata/token_classification.conll.tfrecord")
        print(next(iter(dataset)))

    def test_datapipe_from_jsonl_files(self):
        print()
        dataset = DatapipeForTokenClassification.from_jsonl_files(
            "testdata/token_classification.jsonl",
            feature_vocab_file="testdata/vocab.txt",
            label_vocab_file="testdata/token_classification_vocab.txt",
        )
        print(next(iter(dataset)))

    def test_datapipe_from_conll_files(self):
        print()
        dataset = DatapipeForTokenClassification.from_conll_files(
            "testdata/token_classification.conll.txt",
            feature_vocab_file="testdata/vocab.txt",
            label_vocab_file="testdata/token_classification_vocab.txt",
        )
        print(next(iter(dataset)))


if __name__ == "__main__":
    unittest.main()



from tokenizers import BertWordPieceTokenizer

from .example import ExampleForMaskedLanguageModel
from .masking_strategy import WholeWordMask


class ParserForMaskedLanguageModel:
    """ """

    @classmethod
    def from_vocab_file(cls, vocab_file, **kwargs):
        return cls(tokenizer=None, vocab_file=vocab_file, **kwargs)

    @classmethod
    def from_tokenizer(cls, tokenizer: BertWordPieceTokenizer, **kwargs):
        return cls(tokenizer=tokenizer, vocab_file=None, **kwargs)

    def __init__(self, tokenizer: BertWordPieceTokenizer, vocab_file=None, do_lower_case=True, **kwargs) -> None:
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided!"
        self.tokenizer = tokenizer or BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)
        self.vocabs = list(self.tokenizer.get_vocab().keys())
        self.masking = WholeWordMask(vocabs=self.vocabs, **kwargs)

    def parse(self, instance, max_sequence_length=512, **kwargs) -> ExampleForMaskedLanguageModel:
        sequence = instance["sequence"]
        # set add_special_tokens=False here, masking strategy will add these special tokens
        encoding = self.tokenizer.encode(sequence, add_special_tokens=False)
        masking_results = self.masking(tokens=encoding.tokens, max_sequence_length=max_sequence_length, **kwargs)
        origin_tokens, masked_tokens = masking_results.origin_tokens, masking_results.masked_tokens
        example = ExampleForMaskedLanguageModel(
            tokens=masked_tokens,
            input_ids=[self.tokenizer.token_to_id(x) for x in masked_tokens],
            segment_ids=[0] * len(masked_tokens),
            attention_mask=[1] * len(masked_tokens),
            masked_ids=[self.tokenizer.token_to_id(x) for x in origin_tokens],
            masked_pos=masking_results.masked_indexes,
        )
        return example



import os
from collections import namedtuple

ExampleForQuestionAnswering = namedtuple(
    "ExampleForQuestionAnswering",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "start", "end"],
)



import json
import logging
import os
import re


def read_conll_files(input_files, sep="[\\s\t]+", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning("File %s does not exist, skipped.", f)
            continue
        feature, label = [], []
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    instance = {"feature": feature, "label": label}
                    yield instance
                    feature, label = [], []
                parts = re.split(sep, line)
                if len(parts) != 2:
                    continue
                feature.append(parts[0].strip())
                label.append(parts[1].strip())

        if feature and label:
            instance = {"feature": feature, "label": label}
            yield instance


def read_jsonl_files(input_files, feature_key="feature", label_key="label", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        if not os.path.exists(f):
            logging.warning("File %s does not exist, skipped.", f)
            continue
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {"feature": data[feature_key], "label": data[label_key]}
                yield instance



from .example import ExampleForTokenClassification
from .tokenizers import BertCharLevelTokenizer, LabelTokenizerForTokenClassification


class ParserForTokenClassification:
    """Example parser for token classification"""

    def __init__(
        self,
        feature_tokenizer: BertCharLevelTokenizer = None,
        feature_vocab_file=None,
        label_tokenizer: LabelTokenizerForTokenClassification = None,
        label_vocab_file=None,
        do_lower_case=True,
        o_token="O",
        **kwargs
    ) -> None:
        assert feature_tokenizer or feature_vocab_file, "`feature_tokenizer` or `feature_vocab_file` must be provided!"
        assert label_tokenizer or label_vocab_file, "`label_tokenizer` or `label_vocab_file` must be provided!"

        self.feature_tokenizer = feature_tokenizer or BertCharLevelTokenizer.from_file(
            feature_vocab_file,
            do_lower_case=do_lower_case,
        )
        self.label_tokenizer = label_tokenizer or LabelTokenizerForTokenClassification.from_file(
            label_vocab_file,
            o_token=o_token,
        )

    @classmethod
    def from_vocab_file(cls, feature_vocab_file, label_vocab_file, **kwargs):
        return cls(
            feature_tokenizer=None,
            feature_vocab_file=feature_vocab_file,
            label_tokenizer=None,
            label_vocab_file=label_vocab_file,
            **kwargs
        )

    @classmethod
    def from_tokenizer(
        cls, feature_tokenizer: BertCharLevelTokenizer, label_tokenizer: LabelTokenizerForTokenClassification = None, **kwargs
    ):
        return cls(
            feature_tokenizer=feature_tokenizer,
            feature_vocab_file=None,
            label_tokenizer=label_tokenizer,
            label_vocab_file=None,
            **kwargs
        )

    def parse(self, instance, **kwargs) -> ExampleForTokenClassification:
        feature, label = instance["feature"], instance["label"]
        token_ids = [self.feature_tokenizer.token_to_id(x) for x in feature]
        label_ids = self.label_tokenizer.labels_to_ids(label, add_cls=False, add_sep=False)
        example = ExampleForTokenClassification(
            tokens=feature,
            labels=label,
            input_ids=token_ids,
            segment_ids=[0] * len(token_ids),
            attention_mask=[1] * len(token_ids),
            label_ids=label_ids,
        )
        return example



import abc
import logging
from typing import Dict, List

import tensorflow as tf
from smile_datasets import utils
from smile_datasets.dataset import Datapipe, Dataset
from tokenizers import BertWordPieceTokenizer

from .example import ExampleForQuestionAnswering
from .parsers import ParserForQuestionAnswering
from .readers import read_dureader_checklist, read_dureader_rubost, read_jsonl_files


class DatasetForQuestionAnswering(Dataset):
    """ """

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __getitem__(self, index) -> ExampleForQuestionAnswering:
        raise NotImplementedError()

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples to tfrecord"""

        def _encoding(example: ExampleForQuestionAnswering):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "start": utils.int64_feature([int(example.start)]),
                "end": utils.int64_feature([int(example.end)]),
            }
            return feature

        utils.save_tfrecord(iter(self), _encoding, output_files, **kwargs)


class DatapipeForQuestionAnswering(Datapipe):
    """Datapipe for question answering."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @classmethod
    def from_tfrecord_files(cls, input_files, num_parallel_calls=None, buffer_size=None, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = num_parallel_calls or utils.AUTOTUNE
        buffer_size = buffer_size or utils.AUTOTUNE
        # parse examples
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["start"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["end"])), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_dureader_robust(cls, input_files, tokenizer=None, vocab_file=None, **kwargs) -> tf.data.Dataset:
        instances = read_dureader_rubost(input_files, **kwargs)
        return cls.from_instances(instances, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)

    @classmethod
    def from_dureader_checklist(cls, input_files, tokenizer=None, vocab_file=None, **kwargs) -> tf.data.Dataset:
        instances = read_dureader_checklist(input_files, **kwargs)
        return cls.from_instances(instances, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, tokenizer=None, vocab_file=None, **kwargs) -> tf.data.Dataset:
        instances = read_jsonl_files(input_files, **kwargs)
        return cls.from_instances(instances, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)

    @classmethod
    def from_instances(
        cls, instances: List[Dict], tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ) -> tf.data.Dataset:
        """Build tf.data.Dataset from json instances.

        Args:
            instances: List instance of dict, each instance contains keys `context`, `question`, `answer` and `id`
            tokenizer: Tokenizer used to tokenize text
            vocab_file: The vocab path to build tokenizer. The `tokenizer` or `vocab_file` must be provided!

        Returns:
            Instance of tf.data.Dataset, can be used to fit to tf.keras.Model directly.
        """
        examples = []
        parser = ParserForQuestionAnswering(tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)
        for instance in instances:
            if not instance:
                continue
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        logging.info("Read %d examples in total.", len(examples))
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_dataset(cls, dataset: DatasetForQuestionAnswering, **kwargs) -> tf.data.Dataset:
        examples = []
        for idx in range(len(dataset)):
            e = dataset[idx]
            if not e:
                continue
            examples.append(e)
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(cls, examples: List[ExampleForQuestionAnswering], **kwargs) -> tf.data.Dataset:
        d = cls(**kwargs)

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c, x, y: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, start_key="start", end_key="end", **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, x, y: ((a, b, c), (x, y)),
                num_parallel_calls=num_parallel_calls,
            )
            return dataset
        dataset = dataset.map(
            lambda a, b, c, x, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, {start_key: x, end_key: y}),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(kwargs.get("buffer_size", utils.AUTOTUNE))
        return dataset

    def _batch_padding(self, dataset: tf.data.Dataset, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(kwargs.get("pad_id", 0), dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([None, ], [None, ], [None, ], [], []))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, None, None))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, **kwargs) -> tf.data.Dataset:
        maxlen = tf.constant(kwargs.get("max_sequence_length", 512), dtype=tf.int32)
        pad_id = tf.constant(kwargs.get("pad_id", 0), dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([maxlen, ], [maxlen, ], [maxlen, ], [], []))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, None, None))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _bucket_padding(self, dataset: tf.data.Dataset, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(kwargs.get("pad_id", 0), dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [], [])
        padding_values = (pad_id, pad_id, pad_id, None, None)
        # fmt: on
        dataset = utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, x, y: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )
        return dataset



import abc

import tensorflow as tf


class Dataset(abc.ABC):
    """Abstract dataset"""

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __getitem__(self, index):
        raise NotImplementedError()

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    @abc.abstractmethod
    def save_tfrecord(self, output_files, **kwargs):
        raise NotImplementedError()


class Datapipe(abc.ABC):
    """Abstract dataset"""

    def __init__(self, **kwargs) -> None:
        super().__init__()

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        """Build tf.data.Dataset from tfrecord files."""
        raise NotImplementedError()

    @classmethod
    def from_jsonl_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        """Build tf.data.Dataset from jsonl files."""
        raise NotImplementedError()

    @classmethod
    def from_instances(cls, instances, **kwargs) -> tf.data.Dataset:
        """Build tf.data.Dataset from json instances."""
        raise NotImplementedError()

    @classmethod
    def from_dataset(cls, dataset: Dataset, **kwargs) -> tf.data.Dataset:
        """Build tf.data.Dataset from instances of Dataset."""
        raise NotImplementedError()

    @classmethod
    def from_examples(cls, examples, **kwargs) -> tf.data.Dataset:
        """Build tf.data.Dataset from list of examples."""
        raise NotImplementedError()

    def __call__(
        self,
        dataset: tf.data.Dataset,
        batch_size=32,
        pad_id=0,
        max_sequence_length=512,
        padding_strategy="bucket",
        num_buckets=8,
        bucket_boundaries=[64, 128, 192, 256, 320, 384, 448],
        bucket_batch_sizes=None,
        drop_remainder=False,
        do_filter=True,
        do_repeat=False,
        repeat_count=None,
        do_shuffle=True,
        shuffle_buffer_size=1000000,
        shuffle_seed=None,
        reshuffle_each_iteration=True,
        to_dict=True,
        auto_shard_policy=None,
        **kwargs,
    ) -> tf.data.Dataset:
        dataset = self._filter(dataset, do_filter=do_filter, max_sequence_length=max_sequence_length, **kwargs)
        dataset = self._repeat(dataset, do_repeat=do_repeat, repeat_count=repeat_count, **kwargs)
        dataset = self._shuffle(
            dataset,
            do_shuffle=do_shuffle,
            shuffle_buffer_size=shuffle_buffer_size,
            shuffle_seed=shuffle_seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
            **kwargs,
        )
        dataset = self._padding(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            padding_strategy=padding_strategy,
            max_sequence_length=max_sequence_length,
            num_buckets=num_buckets,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        dataset = self._to_dict(dataset, to_dict=to_dict, **kwargs)
        dataset = self._auto_shard(dataset, auto_shard_policy=auto_shard_policy, **kwargs)
        return dataset

    @abc.abstractmethod
    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        raise NotImplementedError()

    def _repeat(self, dataset: tf.data.Dataset, do_repeat=False, repeat_count=None, **kwargs) -> tf.data.Dataset:
        if not do_repeat:
            return dataset
        if repeat_count is None:
            return dataset
        return dataset.repeat(repeat_count)

    def _shuffle(
        self,
        dataset: tf.data.Dataset,
        do_shuffle=True,
        shuffle_buffer_size=100000,
        shuffle_seed=None,
        reshuffle_each_iteration=True,
        **kwargs,
    ) -> tf.data.Dataset:
        if not do_shuffle:
            return dataset
        dataset = dataset.shuffle(
            buffer_size=shuffle_buffer_size, seed=shuffle_seed, reshuffle_each_iteration=reshuffle_each_iteration
        )
        return dataset

    def _padding(
        self,
        dataset: tf.data.Dataset,
        batch_size=32,
        pad_id=0,
        max_sequence_length=512,
        padding_strategy="bucket",
        bucket_boundaries=[64, 128, 192, 256, 320, 384, 448],
        bucket_batch_sizes=None,
        num_buckets=8,
        drop_remainder=False,
        **kwargs,
    ) -> tf.data.Dataset:
        if padding_strategy == "fixed":
            return self._fixed_padding(
                dataset,
                batch_size=batch_size,
                pad_id=pad_id,
                max_sequence_length=max_sequence_length,
                drop_remainder=drop_remainder,
                **kwargs,
            )
        if padding_strategy == "batch":
            return self._batch_padding(
                dataset,
                batch_size=batch_size,
                pad_id=pad_id,
                drop_remainder=drop_remainder,
                **kwargs,
            )
        return self._bucket_padding(
            dataset,
            batch_size=batch_size,
            pad_id=pad_id,
            max_sequence_length=max_sequence_length,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            num_buckets=num_buckets,
            drop_remainder=drop_remainder,
            **kwargs,
        )

    @abc.abstractmethod
    def _fixed_padding(
        self, dataset: tf.data.Dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs
    ) -> tf.data.Dataset:
        raise NotImplementedError("Fixed padding not supported yet!")

    @abc.abstractmethod
    def _batch_padding(
        self, dataset: tf.data.Dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs
    ) -> tf.data.Dataset:
        raise NotImplementedError("Batch padding is not supported yet!")

    @abc.abstractmethod
    def _bucket_padding(
        self,
        dataset: tf.data.Dataset,
        batch_size=32,
        pad_id=0,
        max_sequence_length=512,
        bucket_boundaries=[64, 128, 192, 256, 320, 384, 448],
        bucket_batch_sizes=None,
        num_buckets=8,
        drop_remainder=False,
        **kwargs,
    ) -> tf.data.Dataset:
        raise NotImplementedError("Bucket padding is not supported yet!")

    @abc.abstractmethod
    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        raise NotImplementedError()

    def _auto_shard(self, dataset: tf.data.Dataset, auto_shard_policy=None, **kwargs) -> tf.data.Dataset:
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset



import logging
from typing import List

import tensorflow as tf
from smile_datasets import utils
from smile_datasets.dataset import Datapipe, Dataset
from smile_datasets.token_classification.parsers import ParserForTokenClassification
from smile_datasets.token_classification.tokenizers import BertCharLevelTokenizer, LabelTokenizerForTokenClassification

from . import readers
from .example import ExampleForTokenClassification


class DatasetForTokenClassification(Dataset):
    """Dataset for token classification task."""

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, index) -> ExampleForTokenClassification:
        raise NotImplementedError()

    def save_tfrecord(self, output_files, **kwargs):
        """Save examples in tfrecord format"""

        def _encode(example: ExampleForTokenClassification):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "label_ids": utils.int64_feature([int(x) for x in example.label_ids]),
            }
            return feature

        utils.save_tfrecord(iter(self), _encode, output_files, **kwargs)


class DatapipeForTokenClassification(Datapipe):
    """Datapipe for token classification"""

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def from_conll_files(
        cls,
        input_files,
        feature_tokenizer: BertCharLevelTokenizer = None,
        feature_vocab_file=None,
        label_tokenizer: LabelTokenizerForTokenClassification = None,
        label_vocab_file=None,
        **kwargs
    ) -> tf.data.Dataset:
        instances = readers.read_conll_files(input_files, **kwargs)
        return cls.from_instances(
            instances,
            feature_tokenizer=feature_tokenizer,
            feature_vocab_file=feature_vocab_file,
            label_tokenizer=label_tokenizer,
            label_vocab_file=label_vocab_file,
            **kwargs,
        )

    @classmethod
    def from_jsonl_files(
        cls,
        input_files,
        feature_tokenizer: BertCharLevelTokenizer = None,
        feature_vocab_file=None,
        label_tokenizer: LabelTokenizerForTokenClassification = None,
        label_vocab_file=None,
        **kwargs
    ) -> tf.data.Dataset:
        instances = readers.read_jsonl_files(input_files, **kwargs)
        return cls.from_instances(
            instances,
            feature_tokenizer=feature_tokenizer,
            feature_vocab_file=feature_vocab_file,
            label_tokenizer=label_tokenizer,
            label_vocab_file=label_vocab_file,
            **kwargs,
        )

    @classmethod
    def from_tfrecord_files(cls, input_files, num_parallel_calls=None, buffer_size=None, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = num_parallel_calls or utils.AUTOTUNE
        buffer_size = buffer_size or utils.AUTOTUNE
        # parse example
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "label_ids": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["label_ids"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_instances(
        cls,
        instances,
        feature_tokenizer: BertCharLevelTokenizer = None,
        feature_vocab_file=None,
        label_tokenizer: LabelTokenizerForTokenClassification = None,
        label_vocab_file=None,
        **kwargs
    ) -> tf.data.Dataset:
        parser = ParserForTokenClassification(
            feature_tokenizer=feature_tokenizer,
            feature_vocab_file=feature_vocab_file,
            label_tokenizer=label_tokenizer,
            label_vocab_file=label_vocab_file,
            **kwargs,
        )
        examples = []
        for instance in instances:
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_dataset(cls, dataset: Dataset, **kwargs) -> tf.data.Dataset:
        examples = []
        for idx in range(len(dataset)):
            e = dataset[idx]
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(cls, examples: List[ExampleForTokenClassification], verbose=True, **kwargs) -> tf.data.Dataset:
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None

        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        # parse examples to dataset
        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.label_ids for e in examples]),
            )
        )
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filer=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filer:
            return dataset
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, y: ((a, b, c), y),
                num_parallel_calls=num_parallel_calls,
            )
            return dataset
        label_key = kwargs.get("label_key", "label_ids")
        dataset = dataset.map(
            lambda a, b, c, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, {label_key: y}),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(kwargs.get("buffer_size", utils.AUTOTUNE))
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([None, ], [None, ], [None, ], [None, ]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.batching_and_padding(dataset, padded_shapes, padding_values, **kwargs)
        return dataset

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = kwargs.get("padded_shapes", ([None, ], [None, ], [None, ], [None, ]))
        padding_values = kwargs.get("padding_values", (pad_id, pad_id, pad_id, pad_id))
        # fmt: on
        dataset = utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, y: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )
        return dataset



import abc
import random
from collections import namedtuple

ResultForMasking = namedtuple("ResultForMasking", ["origin_tokens", "masked_tokens", "masked_indexes"])


class AbstractMaskingStrategy(abc.ABC):
    """Abstract masking strategy"""

    @abc.abstractmethod
    def __call__(self, tokens, **kwargs) -> ResultForMasking:
        raise NotImplementedError()


class WholeWordMask(AbstractMaskingStrategy):
    """Default masking strategy from BERT."""

    def __init__(self, vocabs, change_prob=0.15, mask_prob=0.8, rand_prob=0.1, keep_prob=0.1, max_predictions=20, **kwargs):
        self.vocabs = vocabs
        self.change_prob = change_prob
        self.mask_prob = mask_prob / (mask_prob + rand_prob + keep_prob)
        self.rand_prob = rand_prob / (mask_prob + rand_prob + keep_prob)
        self.keep_prob = keep_prob / (mask_prob + rand_prob + keep_prob)
        self.max_predictions = max_predictions

    def __call__(self, tokens, max_sequence_length=512, **kwargs) -> ResultForMasking:
        tokens = self._truncate_sequence(tokens, max_sequence_length - 2)
        if not tokens:
            return None
        num_to_predict = min(self.max_predictions, max(1, round(self.change_prob * len(tokens))))
        cand_indexes = self._collect_candidates(tokens)
        # copy original tokens
        masked_tokens = [x for x in tokens]
        masked_indexes = [0] * len(tokens)
        for piece_indexes in cand_indexes:
            if sum(masked_indexes) >= num_to_predict:
                break
            if sum(masked_indexes) + len(piece_indexes) > num_to_predict:
                continue
            if any(masked_indexes[idx] == 1 for idx in piece_indexes):
                continue
            for index in piece_indexes:
                masked_indexes[index] = 1
                masked_tokens[index] = self._masking_tokens(index, tokens, self.vocabs)

        # add special tokens
        tokens = ["[CLS]"] + tokens + ["[SEP]"]
        masked_tokens = ["[CLS]"] + masked_tokens + ["[SEP]"]
        masked_indexes = [0] + masked_indexes + [0]
        assert len(tokens) == len(masked_tokens) == len(masked_indexes)
        return ResultForMasking(origin_tokens=tokens, masked_tokens=masked_tokens, masked_indexes=masked_indexes)

    def _masking_tokens(self, index, tokens, vocabs, **kwargs):
        # 80% of the time, replace with [MASK]
        if random.random() < self.mask_prob:
            return "[MASK]"
        # 10% of the time, keep original
        p = self.rand_prob / (self.rand_prob + self.keep_prob)
        if random.random() < p:
            return tokens[index]
        # 10% of the time, replace with random word
        masked_token = vocabs[random.randint(0, len(vocabs) - 1)]
        return masked_token

    def _collect_candidates(self, tokens):
        cand_indexes = [[]]
        for idx, token in enumerate(tokens):
            if cand_indexes and token.startswith("##"):
                cand_indexes[-1].append(idx)
                continue
            cand_indexes.append([idx])
        random.shuffle(cand_indexes)
        return cand_indexes

    def _truncate_sequence(self, tokens, max_tokens=512, **kwargs):
        while len(tokens) > max_tokens:
            if len(tokens) > max_tokens:
                tokens.pop(0)
                # truncate whole world
                while tokens and tokens[0].startswith("##"):
                    tokens.pop(0)
            if len(tokens) > max_tokens:
                while tokens and tokens[-1].startswith("##"):
                    tokens.pop()
                if tokens:
                    tokens.pop()
        return tokens



import json


def read_dureader_rubost(input_files, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for input_file in input_files:
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            info = json.load(fin)
        for d in info["data"]:
            paragraphs = d["paragraphs"]
            for p in paragraphs:
                context = p["context"]
                for qa in p["qas"]:
                    question = qa["question"]
                    _id = qa["id"]
                    if not qa["answers"]:
                        continue
                    answer = qa["answers"][0]["text"]
                    instance = {"context": context, "question": question, "answer": answer, "id": _id}
                    yield instance


def read_dureader_checklist(input_files, **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for input_file in input_files:
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            info = json.load(fin)
        for data in info["data"]:
            for p in data["paragraphs"]:
                title = p["title"]
                context = p["context"]
                for qa in p["qas"]:
                    if qa["is_impossible"]:
                        continue
                    question = qa["question"]
                    answer = qa["answers"][0]["text"]
                    instance = {"context": title + context, "question": question, "answer": answer, "id": qa["id"]}
                    yield instance


def read_jsonl_files(input_files, context_key="context", question_key="question", answers_key="answer", id_key="id", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    _id = 0
    for input_file in input_files:
        with open(input_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                data = json.loads(line)
                answers = data[answers_key]
                if not answers:
                    continue
                if not isinstance(answers, list):
                    answers = [answers]
                answer = answers[0]
                instance_id = data.get(id_key, _id)
                _id += 1
                instance = {"context": data[context_key], "question": data[question_key], "answer": answer, "id": instance_id}
                yield instance



from collections import namedtuple

TokenizerEncoding = namedtuple("TokenizerEncoding", ["text", "tokens", "ids", "type_ids", "attention_mask", "offsets"])


class LabelTokenizerForTokenClassification:
    """Tokenizer for labels, used in token classification tasks."""

    def __init__(self, label2id, o_token="O", **kwargs) -> None:
        self.label2id = label2id
        self.id2label = {v: k for k, v in label2id.items()}
        self.o_token = o_token
        self.o_id = self.label2id[o_token]

    def label_to_id(self, token, **kwargs):
        return self.label2id.get(token, self.o_id)

    def id_to_label(self, id, **kwargs):
        return self.id2label.get(id, self.o_token)

    def labels_to_ids(self, tokens, add_cls=False, add_sep=False, **kwargs):
        ids = [self.label2id.get(token, self.o_id) for token in tokens]
        if add_cls:
            ids = [self.o_id] + ids
        if add_sep:
            ids = ids + [self.o_id]
        return ids

    def ids_to_labels(self, ids, del_cls=True, del_sep=True, **kwargs):
        tokens = [self.id2label.get(_id, self.o_token) for _id in ids]
        if tokens and del_cls and tokens[0] == self.o_token:
            tokens = tokens[1:]
        if tokens and del_sep and tokens[-1] == self.o_token:
            tokens = tokens[:-1]
        return tokens

    @classmethod
    def from_file(cls, vocab_file, o_token="O", **kwargs):
        label2id = cls._read_label_vocab(vocab_file)
        return cls(label2id=label2id, o_token=o_token, **kwargs)

    @classmethod
    def _read_label_vocab(cls, vocab_file):
        m = {}
        with open(vocab_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                k = line.strip()
                m[k] = len(m)
        return m


class BertCharLevelTokenizer:
    """Bert char level tokenizer for token classification tasks."""

    def __init__(
        self,
        token2id,
        do_lower_case=True,
        unk_token="[UNK]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]",
        **kwargs
    ):
        super().__init__()
        self.token2id = token2id
        self.id2token = {v: k for k, v in self.token2id.items()}
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.cls_token = cls_token
        self.sep_token = sep_token
        self.mask_token = mask_token
        self.unk_id = self.token2id[self.unk_token]
        self.pad_id = self.token2id[self.pad_token]
        self.cls_id = self.token2id[self.cls_token]
        self.sep_id = self.token2id[self.sep_token]
        self.mask_id = self.token2id[self.mask_token]
        self.do_lower_case = do_lower_case

    @classmethod
    def from_file(cls, vocab_file, **kwargs):
        token2id = cls._load_vocab(vocab_file=vocab_file)
        tokenizer = cls(token2id, **kwargs)
        return tokenizer

    @classmethod
    def _load_vocab(cls, vocab_file):
        vocab = {}
        idx = 0
        with open(vocab_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                word = line.rstrip("\n")
                vocab[word] = idx
                idx += 1
        return vocab

    def _encode_legacy(self, text, add_cls=True, add_sep=True, **kwargs):
        tokens, ids, type_ids, attention_mask, offsets = [], [], [], [], []
        if self.do_lower_case:
            text = str(text).lower()
        for idx, char in enumerate(text):
            tokens.append(char)
            ids.append(self.token2id.get(char, self.unk_id))
            type_ids.append(0)
            attention_mask.append(1)
            offsets.append((idx, idx + 1))

        if add_cls:
            tokens.insert(0, self.cls_token)
            ids.insert(0, self.cls_id)
            type_ids.insert(0, 0)
            attention_mask.insert(0, 1)
            offsets.insert(0, (0, 0))

        if add_sep:
            tokens.append(self.sep_token)
            ids.append(self.sep_id)
            type_ids.append(0)
            attention_mask.append(1)
            offsets.append((0, 0))

        encoding = TokenizerEncoding(
            text=text,
            tokens=tokens,
            ids=ids,
            type_ids=type_ids,
            attention_mask=attention_mask,
            offsets=offsets,
        )
        return encoding

    def encode(self, sequence, pair=None, add_special_tokens=False, **kwargs):
        """Encode text to encodings"""
        if self.do_lower_case:
            sequence = [x.lower() for x in sequence]
        if not pair:
            encoding = self._encode_sequence(sequence, add_special_tokens=add_special_tokens, **kwargs)
            return encoding
        encoding = self._encode_pair(sequence, pair, add_special_tokens=True, **kwargs)
        return encoding

    def _encode_sequence(self, sequence, add_special_tokens=False, **kwargs):
        encoding = self._encode_legacy(sequence, add_cls=add_special_tokens, add_sep=add_special_tokens, **kwargs)
        return encoding

    def _encode_pair(self, sequence, pair, add_special_tokens=True, **kwargs):
        seq_encoding = self._encode_legacy(sequence, add_cls=True, add_sep=True)
        pai_encoding = self._encode_legacy(pair, add_cls=False, add_sep=True)
        encoding = TokenizerEncoding(
            text=seq_encoding.text + pai_encoding.text,
            tokens=seq_encoding.tokens + pai_encoding.tokens,
            ids=seq_encoding.ids + pai_encoding.ids,
            type_ids=[0] * len(seq_encoding.type_ids) + [1] * len(pai_encoding.type_ids),
            attention_mask=[1] * (len(seq_encoding.attention_mask) + len(pai_encoding.attention_mask)),
            offsets=None,  # offsets set to None in CharLevel tokenizer
        )
        return encoding

    def decode(self, ids, **kwargs):
        tokens = [self.id2token.get(_id, self.unk_token) for _id in ids]
        return tokens

    def token_to_id(self, token, **kwargs):
        if self.do_lower_case:
            token = token.lower()
        return self.token2id.get(token, self.unk_id)

    def id_to_token(self, id, **kwargs):
        return self.id2token.get(id, self.unk_token)



import abc
import json
import logging
import re
from typing import Dict

from tokenizers import BertWordPieceTokenizer

from .example import ExampleForQuestionAnswering


class AbstractExampleParser(abc.ABC):
    """Abc example parser"""

    @abc.abstractmethod
    def parse(self, instance: Dict, **kwargs):
        raise NotImplementedError()


class ParserForQuestionAnswering(AbstractExampleParser):
    """Parse example for qa task"""

    def __init__(self, tokenizer=None, vocab_file=None, do_lower_case=True, **kwargs) -> None:
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided!"
        self.tokenizer = tokenizer or BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)

    def from_vocab(cls, vocab_file, **kwargs):
        tokenizer = BertWordPieceTokenizer.from_file(vocab_file, lowercase=kwargs.get("do_lower_case", True))
        return cls.from_tokenizer(tokenizer, **kwargs)

    def from_tokenizer(cls, tokenizer: BertWordPieceTokenizer, **kwargs):
        return cls(tokenizer=tokenizer, vocab_file=None, **kwargs)

    def _find_answer_span(self, context, answer):
        for m in re.finditer(re.escape(answer), context, re.IGNORECASE):
            start, end = m.span()
            return start, end
        return 0, 0

    def parse(self, instance: Dict, **kwargs) -> ExampleForQuestionAnswering:
        context, question, answer = instance["context"], instance["question"], instance["answer"]
        start_char_idx, end_char_idx = self._find_answer_span(context, answer)
        if end_char_idx <= start_char_idx:
            return None
        # Mark the character indexes in context that are in answer
        is_char_in_ans = [0] * len(context)
        for idx in range(start_char_idx, end_char_idx):
            is_char_in_ans[idx] = 1
        context_encoding = self.tokenizer.encode(context, add_special_tokens=True)
        # Find tokens that were created from answer characters
        ans_token_idx = []
        for idx, (start_char_idx, end_char_idx) in enumerate(context_encoding.offsets):
            if sum(is_char_in_ans[start_char_idx:end_char_idx]) > 0:
                ans_token_idx.append(idx)
        if not ans_token_idx:
            logging.warning("No answer instance: %s", json.dumps(instance, ensure_ascii=False))
            logging.warning("Skipped.")
            return None
        start_token_idx, end_token_idx = ans_token_idx[0], ans_token_idx[-1]

        question_encoding = self.tokenizer.encode(question, add_special_tokens=True)
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(question_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        assert len(input_ids) == len(segment_ids), "input_ids length:{} VS segment_ids length: {}".format(
            len(input_ids), len(segment_ids)
        )
        assert len(input_ids) == len(attention_mask), "input_ids length:{} VS attention_mask length: {}".format(
            len(input_ids), len(attention_mask)
        )
        tokens = context_encoding.tokens + question_encoding.tokens[1:]
        valid = self.validate_answer(tokens, start_token_idx, end_token_idx, answer)
        if not valid:
            logging.warning("Invalid instance: %s", json.dumps(instance, ensure_ascii=False))
            logging.warning("Skipped.")
            return None
        example = ExampleForQuestionAnswering(
            tokens=tokens,
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=start_token_idx,
            end=end_token_idx,
        )
        return example

    def validate_answer(self, tokens, start_index, end_index, answer):
        words = []
        for idx in range(start_index, end_index + 1):
            w = tokens[idx].lstrip("##")
            if w == "[UNK]":
                logging.warning("answer tokens constains [UNK]: start_index=%d, end_index=%d", start_index, end_index)
                return False
            words.append(w)
        words = "".join(words).lower()

        def _normalize(x):
            x = x.lower()
            x = "".join(x.split())
            return x

        words = _normalize(words)
        answer = _normalize(answer)
        if words != answer:
            logging.warning("tokens: {} != answer: {}".format(words, answer))
            return False
        return True



import json
import logging
import os


def read_jsonl_files(input_files, sequence_key="sequence", **kwargs):
    if isinstance(input_files, str):
        input_files = [input_files]
    for f in input_files:
        with open(f, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                data = json.loads(line)
                instance = {"sequence": data[sequence_key]}
                yield instance



import abc
import logging
from enum import auto
from typing import List

import tensorflow as tf
from smile_datasets import utils
from smile_datasets.dataset import Datapipe, Dataset
from smile_datasets.simcse.parsers import ParserForSupervisedSimCSE
from tokenizers import BertWordPieceTokenizer

from . import readers
from .example import ExampleForSupervisedSimCSE


class DatasetForSupervisedSimCSE(Dataset):
    """dataset for unsupervised SimCSE"""

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, index) -> ExampleForSupervisedSimCSE:
        raise NotImplementedError()

    def save_tfrecord(self, output_files, **kwargs):
        """Convert examples to tfrecord"""

        def _encoding(example: ExampleForSupervisedSimCSE):
            feature = {
                "input_ids": utils.int64_feature([int(x) for x in example.input_ids]),
                "segment_ids": utils.int64_feature([int(x) for x in example.segment_ids]),
                "attention_mask": utils.int64_feature([int(x) for x in example.attention_mask]),
                "pos_input_ids": utils.int64_feature([int(x) for x in example.pos_input_ids]),
                "pos_segment_ids": utils.int64_feature([int(x) for x in example.pos_segment_ids]),
                "pos_attention_mask": utils.int64_feature([int(x) for x in example.pos_attention_mask]),
            }
            return feature

        utils.save_tfrecord(iter(self), _encoding, output_files, **kwargs)


class DatapipeForSupervisedSimCSE(Datapipe):
    """Datapipe for unsupervised SimCSE"""

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        dataset = utils.read_tfrecord_files(input_files, **kwargs)
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        # parse example
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "pos_input_ids": tf.io.VarLenFeature(tf.int64),
            "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
            "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    @classmethod
    def from_jsonl_files(cls, input_files, **kwargs) -> tf.data.Dataset:
        instances = readers.read_jsonl_files_for_supervised_simcse(input_files, **kwargs)
        return cls.from_instances(instances, **kwargs)

    @classmethod
    def from_instances(cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs) -> tf.data.Dataset:
        parser = ParserForSupervisedSimCSE(tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)
        examples = []
        for instance in instances:
            if not instance:
                continue
            e = parser.parse(instance, **kwargs)
            if not e:
                continue
            examples.append(e)
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_dataset(cls, dataset: Dataset, **kwargs) -> tf.data.Dataset:
        examples = [e for _, e in enumerate(dataset) if e]
        return cls.from_examples(examples, **kwargs)

    @classmethod
    def from_examples(cls, examples: List[ExampleForSupervisedSimCSE], verbose=True, **kwargs) -> tf.data.Dataset:
        if not examples:
            logging.warning("examples is empty or null, skipped to build dataset.")
            return None
        if verbose:
            n = min(5, len(examples))
            for i in range(n):
                logging.info("Showing NO.%d example: %s", i, examples[i])

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_attention_mask for e in examples], dtype=tf.int32),
            )
        )
        # do transformation
        d = cls(**kwargs)
        return d(dataset, **kwargs)

    def _filter(self, dataset: tf.data.Dataset, do_filter=True, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        if not do_filter:
            return dataset
        dataset = dataset.filter(lambda a, b, c, d, e, f: tf.size(a) <= max_sequence_length)
        return dataset

    def _to_dict(self, dataset: tf.data.Dataset, to_dict=True, **kwargs) -> tf.data.Dataset:
        num_parallel_calls = kwargs.get("num_parallel_calls", utils.AUTOTUNE)
        buffer_size = kwargs.get("buffer_size", utils.AUTOTUNE)
        if not to_dict:
            dataset = dataset.map(
                lambda a, b, c, d, e, f: ((a, b, c, d, e, f), None),
                num_parallel_calls=num_parallel_calls,
            ).prefetch(buffer_size)
            return dataset
        dataset = dataset.map(
            lambda a, b, c, d, e, f: (
                {
                    "input_ids": a,
                    "segment_ids": b,
                    "attention_mask": c,
                    "pos_input_ids": d,
                    "pos_segment_ids": e,
                    "pos_attention_mask": f,
                },
                None,
            ),
            num_parallel_calls=num_parallel_calls,
        ).prefetch(buffer_size)
        return dataset

    def _fixed_padding(self, dataset: tf.data.Dataset, pad_id=0, max_sequence_length=512, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _batch_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.batching_and_padding(dataset, padded_shapes=padded_shapes, padding_values=padding_values, **kwargs)

    def _bucket_padding(self, dataset: tf.data.Dataset, pad_id=0, **kwargs) -> tf.data.Dataset:
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        padded_shapes = ([None, ], [None, ], [None, ], [None, ], [None, ], [None, ])
        padding_values = (pad_id, pad_id, pad_id, pad_id, pad_id, pad_id)
        # fmt: on
        return utils.bucketing_and_padding(
            dataset,
            bucket_fn=lambda a, b, c, d, e, f: tf.size(a),
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            **kwargs,
        )



from collections import namedtuple

ExampleForUnsupervisedSimCSE = namedtuple(
    "ExampleForUnsupervisedSimCSE",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
    ],
)
ExampleForSupervisedSimCSE = namedtuple(
    "ExampleForSupervisedSimCSE",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
        "pos_tokens",
        "pos_input_ids",
        "pos_segment_ids",
        "pos_attention_mask",
    ],
)

ExampleForHardNegativeSimCSE = namedtuple(
    "ExampleForHardNegativeSimCSE",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
        "pos_tokens",
        "pos_input_ids",
        "pos_segment_ids",
        "pos_attention_mask",
        "neg_tokens",
        "neg_input_ids",
        "neg_segment_ids",
        "neg_attention_mask",
    ],
)



from tokenizers import BertWordPieceTokenizer

from .example import ExampleForSequenceClassification


class ParserForSequenceClassification:
    """ """

    @classmethod
    def from_tokenizer(cls, tokenizer: BertWordPieceTokenizer, **kwargs):
        return cls(tokenizer=tokenizer, vocab_file=None, **kwargs)

    @classmethod
    def from_vocab_file(cls, vocab_file, **kwargs):
        return cls(tokenizers=None, vocab_file=vocab_file, **kwargs)

    def __init__(self, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, do_lower_case=True, **kwargs) -> None:
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided!"
        self.tokenizer = tokenizer or BertWordPieceTokenizer.from_file(vocab_file, lowercase=do_lower_case)

    def parse(self, instance, add_special_tokens=True, **kwargs) -> ExampleForSequenceClassification:
        sequence = instance["sequence"]
        pair = instance.get("pair", None)
        label = int(instance["label"])
        encoding = self.tokenizer.encode(sequence, pair=pair, add_special_tokens=add_special_tokens)
        example = ExampleForSequenceClassification(
            tokens=encoding.tokens,
            input_ids=encoding.ids,
            segment_ids=encoding.type_ids,
            attention_mask=encoding.attention_mask,
            label=label,
        )
        return example
